/* $Id: PitchAndI.java,v 1.65 2007/11/02 12:27:15 mit Exp $ */
/* PitchAndI.java v0.8 */
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.sound.sampled.*;
import java.text.DecimalFormat;

public class PitchAndI extends JPanel 
    implements ActionListener, ChangeListener, MouseListener, 
	       MouseMotionListener, Runnable
{
    private JButton jbOn;
    private JButton[] jbSelect;
    private JButton jbMWide, jbPWide;
    private JButton jbMLevel, jbPLevel;
    private JButton jbAuto;
    private JButton jbMOct, jbPOct;
    private JButton jbMKey, jbPKey;
    private JButton jbLoad;
    private JButton jbClear;
    private JButton jbCut;
    private JComboBox jcPitch;
    private JComboBox jcPart;
    private JComboBox jcTime;
    private JTextField jtForm;
    private JTextField jtFile;
    private JTextField jtKey;
    private JTextField jtData;
    private JTextField jtParf;
    private JTextField jtPkey;
    private JSlider jsLevel;

    private Recording REC;
    private PaintGraph PG;

    private final double PI2 = 2*Math.PI;
    private final int Oct = 12;
    private final int Cent = 1200;
    private final int KB = 88;
    private final int KC = KB+Oct*3+40; // by Part+40
    private final int WAV = 0;
    private final int FFT = 1;
    private final int TUN = 2;
    private final int PIC = 3;
    private final float RATE = 44100.0f;
    private final double ML = Math.log(2)/Cent/Math.log(10);
    private final double MC = Cent/Math.log(2);
    private final int[] Pitchs = {
	415, 435,436,437,438,439,440,441,442,443,444,445};
    private final int[] Time = {1,2,3};
    private final int[] Part = {0,12,19,24,28,31,34,36,38,40};
    private final String[] Selects = {"Wave", "FFT", "Tune", "Ave."};

    private Peak[] K;
    private Peak[] Pic;
    private Peak PK;

    private boolean Onoff;
    private boolean Auto;
    private boolean OLDPK;
    private boolean NOWPK;
    private boolean EXTSW; // ppd/icd
    private boolean CUT;   // head/tail
    private boolean[] Psw; // picupǡonoff sw.
    private int NB;
    private int NN;
    private int N2;
    private int Splngf;
    private int GRP = WAV;
    private int Skey = 49-1; // 򥭡
    private int Pitch = 440;
    private int PART = 1;
    private int Loop;
    private int PTime = 2;
    private int DEVID;     // PTime*RATE/NN
    private int mX, mY;
    private int firstX, lastX;
    private int[] Temp;
    private double DELF;   // Splngf/NN
    private double Level = 50000;
    private double RN;
    private double[] YY;
    private double[] Re;
    private double[] Im;
    private double[] Han;
    private double[] Keyc;
    private double[] Keyi;
    private double[] Data;
    private String Error;

    private DecimalFormat df1;
    private File file;
    private FFT fftN;
    private Thread thr;

    public PitchAndI() {
	PG = new PaintGraph();
	buildLayout();

	Border border = new BevelBorder(BevelBorder.RAISED);
        String name = "Java Petit Pitch";
        border = new TitledBorder(border, name);
        setBorder(border);

	set_sampling();
	initStart();
	set_devid();
	df1 = new DecimalFormat("#0.0##");
    }

    public static void main(String[] args) {
	PitchAndI pi = new PitchAndI();

	JFrame f = new JFrame();
        f.getContentPane().add(pi);
	f.pack();
	f.setSize(500, 460);
	f.setVisible(true);
	f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	f.show();
    }

    public void buildLayout() {
	setLayout(new BorderLayout());

	JPanel jn = new JPanel();
	jn.setLayout(new FlowLayout());

	jbOn = new JButton("OFF");
	jbOn.addActionListener(this);
	jn.add(jbOn);

	jn.add(new JLabel("Time:"));
	
	Integer tim[] = new Integer[Time.length];
        for (int i = 0; i < Time.length; i++)
	    tim[i] = new Integer(Time[i]);
	jcTime = new JComboBox(tim);
      	jcTime.setSelectedIndex(1); // 2sec
	jcTime.addActionListener(this);
        jn.add(jcTime);

	jbCut = new JButton("Head");
	jbCut.addActionListener(this);
	jn.add(jbCut);

        jtFile = new JTextField(10);
        jtFile.setEditable(true);
        jn.add(jtFile);

	jbLoad = new JButton("Load");
        jbLoad.addActionListener(this);
        jn.add(jbLoad);

	jbClear = new JButton("AC");
        jbClear.addActionListener(this);
        jn.add(jbClear);

	Border bo = new BevelBorder(BevelBorder.RAISED);
	jn.setBorder(bo);

	add(jn, "North");
	//-------------

	JPanel js = new JPanel();
	js.setLayout(new GridLayout(4,1,0,0));

	JPanel js0 = new JPanel();
	js0.setLayout(new FlowLayout());

	jtForm = new JTextField(35);
	jtForm.setEditable(false);
	js0.add(jtForm);

	js.add(js0);
	
	JPanel js1 = new JPanel();
	js1.setLayout(new FlowLayout());

	jbAuto = new JButton("Auto.");
	jbAuto.addActionListener(this);
	js1.add(jbAuto);

	jtKey = new JTextField(4);
        jtKey.setEditable(false);
        js1.add(jtKey);

	jbMOct = new JButton("<<");
	jbMOct.addActionListener(this);
	js1.add(jbMOct);

	jbMKey = new JButton("<");
	jbMKey.addActionListener(this);
	js1.add(jbMKey);

	jbPKey = new JButton(">");
	jbPKey.addActionListener(this);
	js1.add(jbPKey);

	jbPOct = new JButton(">>");
	jbPOct.addActionListener(this);
	js1.add(jbPOct);

	js1.add(new JLabel("|"));

	jtPkey = new JTextField(5);
        jtPkey.setEditable(false);
        js1.add(jtPkey);

	js.add(js1);

	JPanel js2 = new JPanel();
	js2.setLayout(new FlowLayout());

	jtData = new JTextField(22);
        jtData.setEditable(false);
        js2.add(jtData);

	js2.add(new JLabel("| Part."));

	Integer prt[] = new Integer[Part.length];
        for (int i = 0; i < Part.length; i++)
            prt[i] = new Integer(i+1);
        jcPart = new JComboBox(prt);
        jcPart.addActionListener(this);
        js2.add(jcPart);

	jtParf = new JTextField(9);
        jtParf.setEditable(false);
        js2.add(jtParf);

	js.add(js2);

	jsLevel = new JSlider(0, 500, (int)(Level/1000));
	jsLevel.setPaintTicks(true);
        jsLevel.setPaintLabels(true);
        jsLevel.setMajorTickSpacing(50);
        jsLevel.setMinorTickSpacing(10);
        jsLevel.addChangeListener(this);
        js.add(jsLevel);

	add(js, "South");
	js.setBorder(bo);
	//-------------

	JPanel je = new JPanel();
	je.setLayout(new GridLayout(5,1,0,0));

	jbPLevel = new JButton("+");
	jbPLevel.addActionListener(this);
	je.add(jbPLevel);

	jbMLevel = new JButton("-");
	jbMLevel.addActionListener(this);
	je.add(jbMLevel);

	je.add(new JLabel(""));

	jbPWide = new JButton(">");
	jbPWide.addActionListener(this);
	je.add(jbPWide);

	jbMWide = new JButton("<");
	jbMWide.addActionListener(this);
	je.add(jbMWide);

	add(je, "East");
	je.setBorder(bo);
	//-------------

	JPanel jw = new JPanel();
	jw.setLayout(new GridLayout(5,1,0,0));

        jbSelect = new JButton[Selects.length];
        for (int i = 0; i < Selects.length; i++) {
            jbSelect[i] = new JButton(Selects[i]);
            jbSelect[i].addActionListener(this);
            jw.add(jbSelect[i]);
        }

	Integer pit[] = new Integer[Pitchs.length];
        for (int i = 0; i < Pitchs.length; i++)
	    pit[i] = new Integer(Pitchs[i]);
	jcPitch = new JComboBox(pit);
	jcPitch.setSelectedIndex(6); // 440
	jcPitch.addActionListener(this);
        jw.add(jcPitch);

	add(jw, "West");
	jw.setBorder(bo);
	//------------

	add(PG, "Center");

	PG.addMouseListener(this);
        PG.addMouseMotionListener(this);

	set_enabled();
    }
    public void set_enabled() {
	jbOn.setText(Onoff ? " ON" : "OFF");

	boolean gf = (GRP == FFT);
	boolean gt = (GRP != TUN);
	jbPWide.setEnabled(gf);
	jbMWide.setEnabled(gf);
	jbPLevel.setEnabled(gt);
	jbMLevel.setEnabled(gt);
	jbMOct.setEnabled(Auto);
	jbMKey.setEnabled(Auto);
	jbPKey.setEnabled(Auto);
	jbPOct.setEnabled(Auto);
	jcTime.setEnabled(!Onoff);
    }
    public void set_sampling() {
	NB = 11;       // n (9:512 10:1024 11:2048 12:4096)
	NN = (int)Math.pow(2, NB); // 2^n
	N2 = NN/2;
	Temp = new int[NN];
	RN = 1.0/NN;
	YY = new double[N2];
	Re = new double[NN];
	Im = new double[NN];
	fftN = new FFT(NN);
	initHanning();
    }

    public void mousePressed(MouseEvent m) {
	if (GRP == PIC)
	    firstX = m.getX();
    }
    public void mouseReleased(MouseEvent m) {
	if (GRP == PIC) {
	    lastX = m.getX();
	    if (lastX < firstX) {
		int tmp = lastX;
		lastX = firstX;
		firstX = tmp;
	    }
	    PG.repaint();
	}
    }
    public void mouseDragged(MouseEvent m) {}
    public void mouseMoved(MouseEvent m) {}
    public void mouseEntered(MouseEvent m) {}
    public void mouseExited(MouseEvent m) {}
    public void mouseClicked(MouseEvent m) {}
    public void stateChanged(ChangeEvent e) {
        if (e.getSource() == jsLevel)
	    Level = (10+jsLevel.getValue())*1000;
    }
    public void actionPerformed(ActionEvent e){
        for (int i = 0; i < jbSelect.length; i++) {
            if (e.getSource() == jbSelect[i])
                GRP = i;
        }

	if (e.getSource() == jbOn) // on/off
	    on_off();

	if (e.getSource() == jbLoad)
	    file_set();
	if (e.getSource() == jbClear)
	    set_clear();

	if (e.getSource() == jbMWide)
	    PG.set_wide(false);
	if (e.getSource() == jbPWide)
	    PG.set_wide(true);
	if (e.getSource() == jbMLevel)
	    PG.set_level(false);
	if (e.getSource() == jbPLevel)
	    PG.set_level(true);

	if (e.getSource() == jcPitch)
	    select_pitch();

	if (e.getSource() == jcPart)
	    select_partial();

	if (e.getSource() == jcTime)
	    set_time();

	if (e.getSource() == jbAuto)
	    set_auto();

	if (e.getSource() == jbMOct)
	    select_key(-Oct);
	if (e.getSource() == jbMKey)
	    select_key(-1);
	if (e.getSource() == jbPKey)
	    select_key(1);
	if (e.getSource() == jbPOct)
	    select_key(Oct);

	if (e.getSource() == jbCut)
	    set_cutout();

	PG.repaint();
	write_err();
	set_enabled();
    }
    public void set_cutout() {
	CUT = !CUT;

	jbCut.setText(CUT ? "Tail" : "Head");
    }
    public void select_partial() {
	PART = ((Integer)(jcPart.getSelectedItem())).intValue();
    }
    public void set_clear() {
	initc(0);
	init_inh(0);
	jtFile.setText("");
	file = null;
	select_pitch();
    }
    public void file_set() {
	Error = null;
        file_open();
        if (file != null) {
            write_filenm(file);
            data_in();
	    add_cent(get_pitch());
        }
    }
    public void data_in() {
        int n, max = Data.length/2;

        for (int i = 0; i < max; i++) {
            n = i*2;
	    if (EXTSW) // icd
		Keyi[(int)Data[n]-1] = Data[n+1];
	    else       // ppd  
		Keyc[(int)Data[n]-1] = Data[n+1];
        }
    }
    public void write_filenm(File fn) {
        jtFile.setText(fn.getName());
    }
    public void file_open() {
	JFileChooser fc = null;
	try {
            File fi = new File(System.getProperty("user.dir"));
            fc = new JFileChooser(fi);
            fc.setFileFilter(new javax.swing.filechooser.FileFilter() {
                    public boolean accept(File f) {
                        if (f.isDirectory())
                            return true;
                        String name = f.getName();
			if (name.endsWith(".ppd") ||
			    name.endsWith(".icd")) {
			    return true;
			}
			return false;
                    }
                    public String getDescription() {
			return ".ppd, .icd";
                    }
                });
	    if (fc.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
                file = fc.getSelectedFile();
                String nm = file.getName();
		
		if (nm.endsWith(".ppd")) {
		    Keyc = new double[KC]; //
		    EXTSW = false;
		} else if (nm.endsWith(".icd")) {
		    Keyi = new double[KC]; //
		    EXTSW = true;
		}
				    
		try {
		    Data = ReadData.readData(nm); // ReadData
		} catch(Exception e) {
		    Error = "File read error.";
		}
	    }
	} catch (SecurityException ex) {
            Error = "File security error.";
        } catch (Exception ex) {
            Error = "File error.";
        }
    }
    public void set_time() {
	PTime = ((Integer)(jcTime.getSelectedItem())).intValue();
	set_devid();
	reset_pic();
    }
    public void set_devid() {
	DEVID = (int)(PTime*RATE/NN);
	reset_peak();
    }
    public void set_auto() {
	Auto = !Auto;
	jbAuto.setText(Auto ? "Manu." : "Auto.");
    }
    public void select_key(int num) {
	Skey += num;
	if (Skey < 0)
	    Skey = KB+Skey;
	else if (Skey >= KB)
	    Skey = Skey-KB;
    }
    public double get_pitch() {
	Pitch = ((Integer)(jcPitch.getSelectedItem())).intValue();
	int a49 = 49-1;

	return ftoc(a49, Pitch);
    }
    public void select_pitch() {
	if (file != null) {
	    data_in();
	    add_cent(get_pitch());
	} else
	    initc(get_pitch());
    }
    public void on_off() {
	Onoff = !Onoff;

	if (Onoff) {
	    Splngf = (int)RATE;
	    REC = new Recording(NN, RATE);
	    REC.start();

	    thread_start();

	    write_format(REC.getFormat());
	} else {
	    REC.stop();
	    thread_stop();
	}
    }
    public void thread_start() {
        thr = new Thread(this);
        thr.start();
    }
    public void thread_stop() {
        if (thr != null)
            thr = null;
    }
    public void write_format(AudioFormat f) {
	jtForm.setText(""+f);
    }
    public void write_err() {
	if (Error != null)
	    jtForm.setText("error:"+Error);
    }
    //------------------------

    public void run() {
	int stime = 80;//150;

	try {
	    reset_pic();
	    NOWPK = OLDPK = false;
	    int num = PG.get_select_keynm(); // ?

            while (REC.isRun()) {
		Temp = REC.getData();

		if (GRP != WAV) {
		    fft(Temp);
		    if (GRP == TUN) {
			if (isPeak())
			    PK = pick(num);
		    } else if (GRP == PIC) {
			if (isRaise(isPeak()))
			    reset_pic();
			if (!NOWPK)   //
			    continue;

			pickup(num);
			if (Loop < DEVID)
			    Loop++;
			else if (CUT) //
			    Loop = DEVID-1;
		    }
		}
		PG.repaint();
		thr.sleep(stime);
            }
	} catch (InterruptedException e) {
            Error = "Run error.";
        }
    }
    public boolean isRaise(boolean sw) {
	OLDPK = NOWPK; // 1¸
	NOWPK = sw;    // ߤisPeak
	return (OLDPK == false && NOWPK == true) ? true : false;
    }
    //--------------------------------
 
    public class PaintGraph extends JPanel {
	final int fnt = 16;
	final Color plus = new Color(225, 59, 164);
	final Color minus = new Color(67, 166, 225);
	final Color white = new Color(245, 235, 245);
	private final String[] Keynm = 
	{"A","B","H","C","Cis","D","Es","E","F","Fis","G","Gis"};

	private int W, H, H2, W2;
	private int Xby = 2;  // fft
	private int Cby = 50; // pic
	private int splf;     // NN/Xby
	private int slgf;     // Splngf/Xby
	private int fwd;      // fft 
	private int ri, rj;   // tunȾ
	private double Yby = 0.0003; // fft
	private double Wby = 0.01;   // wav
	private double ds;    // NN
	private double df;    // Splngf
	private double dw;    // temp(wave)
	private double dp;    // pic
	private double di, dj, dc; // tun
	
	private Font FNT = new Font("serif", Font.PLAIN, fnt);
	private FontMetrics fm = getFontMetrics(FNT);
	
	public PaintGraph() {
	    setBackground(Color.black);
	}
	public void paintComponent(Graphics g) {
	    Dimension d = getSize();
	    W = d.width;
	    H = d.height;	
	    H2 = H/2;
	    W2 = W/2;
	    ri = W/10;    // Ⱦ
	    rj = W/8;     //  "
	    di = W2/40.0; // cent
	    dj = W2/9.0;  //  "
	    dc = ri/8.0;  // ѲΨ(8octʬ?)
	    int h0 = H-10;

	    super.paintComponent(g);
	    set_const();
	    g.setFont(FNT);
	    write_key();

	    switch (GRP) {
	    case WAV:
		disp_wave(g);
		break;
	    case FFT:
		disp_memori(g, h0);
		disp_fft(g, h0);
		break;
	    case TUN:
		disp_tuning(g);
		break;
	    case PIC:
		drag_check();
		disp_pickup(g);
		break;
	    default:
	    }
	}
	public void set_wide(boolean sw){
	    Xby = sw ? Xby/2 : Xby*2;
	    if (Xby < 2)
		Xby = 2;

	    set_wh();
	}
	public void set_const() {
	    splf = NN/Xby;
	    slgf = Splngf/Xby;
	    DELF = (double)Splngf/NN;

	    set_wh();
	}
	public void set_wh() {
	    ds = (W-1)/(double)splf;
	    df = (W-1)/(double)slgf;
	    dw = (W-1)/(double)NN;
	    dp = (W-1)/(double)DEVID;

	    set_fft_wide();
	}
	public void set_level(boolean sw) {
	    if (GRP == WAV)
		Wby = sw ? Wby*2 : Wby/2;
	    else if (GRP == FFT)
		Yby = sw ? Yby*2 : Yby/2;
	    else if (GRP == PIC) {
		int max = 10000;
		Cby = sw ? (is5(Cby) ? Cby*2 : Cby*5) :
		    (is5(Cby) ? Cby/5 : Cby/2);

		if (Cby < 5) // 5 ?
		    Cby = 5;
		else if (Cby > max)
		    Cby = max;
	    }
	}
	public boolean is5(int num) {
	    int i = 1;
	    int max = 10000;
	    while (max > i) {
		if (num == 5*i)
		    return true;
		else
		    i *= 10;
	    }

	    return false;
	}
	public void set_fft_wide() {
	    fwd = (ds < 1) ? 1 : (int)ds;
	}
	public void disp_fft(Graphics g, int v0) {
	    double v1;

	    g.setColor(Color.green);
	    for (int i = 0; i < splf; i++) {
		v1 = YY[i]*Yby;
		g.drawRect((int)(ds*i), (int)(v0-v1), fwd, (int)v1);
	    }

	    int lev = (int)(v0-Level*Yby);
	    g.setColor(Color.pink);
	    g.drawLine(0, lev, W-1, lev);  // disp level
	}
	public void disp_memori(Graphics g, int v0) {
	    int spi, l;
	    g.setColor(Color.orange);
	    for (int i = 0; i < slgf; i += 10) {
		if (i % 10000 == 0)
		    l = 10;
		else if (i % 1000 == 0)
		    l = 7;
		else if (i % 500 == 0)
		    l = 4;
		else if (i % 100 == 0)
		    l = 2;
		else
		    continue;
		spi = (int)(df*i);
		g.drawLine(spi, v0, spi, v0+l);
	    }
	}

	public void disp_wave(Graphics g) {
	    g.setColor(Color.green);
	    for (int i = 0; i < NN; i++)
		g.drawOval((int)(dw*i), (int)(H2-Temp[i]*Wby), 1,1); 
	}
	public void disp_tuning(Graphics g) {
	    int h2 = H/2;
	    g.setColor(Color.gray);
	    g.drawLine(0, h2, W-1, h2);
	    g.drawLine(W2, 0, W2, H-1);

	    if (PK == null)
		return;

	    int skey = get_key();
	    double freq = PK.getPeakf();
	    double ic = get_ipcent(skey);
	    double cent = ftoc(get_pkey(skey), freq)-ic;
	    double cen = cent2(cent);
	    int y = h2;
	    int x = cent2x(cen);
	    int r = cent2r(cent, cen);
	    int l = 2*r;
	    g.setColor(select_color(cent));
	    g.fillOval(x-r, y-r, l, l);

	    disp_keyfc(g, skey, freq);
	}
	public void write_key() {
	    int key = get_key();

	    jtKey.setText(get_keyname(key));

	    StringBuffer str = new StringBuffer();
	    str.append(dform(ctof(getFreq(key), Keyc[key]))+"[Hz]:");
	    str.append(dform(Keyc[key])+"[cent]:");
	    str.append(dform(Keyi[key])+"[Ic]");
	    jtData.setText(str.toString());

	    jtParf.setText(dform(get_ipfreq(key))+"[Hz]");
	}

	public double get_ipfreq(int key) {
	    jtPkey.setText(get_keyname(get_pkey(key)));

	    return ctof(getFreq(key)*PART, get_ipcent(key));
	}
	public double get_ipcent(int key) {
	    return get_icent(key)+Keyc[key];
	}
	public double get_icent(int key) {
	    return PART*PART*Keyi[key];
	}
	public int get_pkey(int key) {
	    return key+Part[PART-1];
	}
	public int get_key() {
	    return (Auto) ? Skey :
		(K[0] == null) ? Skey : K[0].getPeakk();
	}
	public String get_keyname(int key) {
	    return (key+1)+Keynm[key%Oct];
	}
	public void disp_keyfc(Graphics g, int key, double freq) {
	    StringBuffer str = new StringBuffer();
	    str.append(dform(freq)+"[Hz]:");
	    str.append(dform(ftoc(get_pkey(key), freq)-get_ipcent(key)));
	    str.append("[cent]");

	    jtForm.setText(str.toString());
	}
	public int cent2x(double cen) {
	    if (cen > 9)          // 10~50
		return (int)(W-(cen-10)*di);
	    else if (cen < -9)    // -10~-50
		return (int)(-(cen+10)*di);
	    else                  // 9~-9
		return (int)(W2+cen*dj);
	}
	public int cent2r(double cent, double cen) {
	    if (cent > 50 || cent < -50) {
		int oc = (int)(Math.abs(cent)/Cent)+1;
		ri -= oc*dc;
		rj -= oc*dc;
		if (ri < 1)
		    ri = 1;
		if (rj < 1)
		    rj = 1;
	    }
	    if (cen > 9 || cen < -9) // -50~50
		return ri;
	    else                     // 9~-9
		return rj;
	}
	public double cent2(double cent) {
	    cent %= 100;

	    if (cent > 50)
		cent -= 100;
	    else if (cent < -50)
		cent += 100;

	    return cent;
	}
	public Color select_color(double cent) {
	    if (cent > 50 || cent < -50)
		return Color.gray;
	    else if (cent > 9)
		return plus;
	    else if (cent < -9)
		return minus;
	    else 
		return Color.white;
	}
	public int get_select_keynm() {
	    if (K[0] == null)
		return 0;

	    int key = get_pkey(Auto ? Skey : K[0].getPeakk());
	    for (int i = 0; i < K[0].getPeakt(); i++) {
		if (key == K[i].getPeakk())
		    return i;
	    }

	    return 0;
	}

	public void disp_cmemori(Graphics g) {
	    g.setColor(Color.gray);
	    g.drawLine(0, H2, W-1, H2);

	    int max = 5, di, sf = fnt-3;
	    int dy = Cby/max;
	    int sl;
	    String ss;
	    double dh = H2/(double)Cby;
	    for (int i = 1; i <= max; i++) {
		di = (int)(H2+dy*dh*i);
		g.setColor(Color.blue);
		g.drawLine(0, di, W-1, di);
		
		di = (int)(H2-dy*dh*i);
		g.drawLine(0, di, W-1, di);

		ss = dy*i+"";
		sl = fm.stringWidth(ss)+2;
		g.setColor(Color.orange);
		g.drawString(ss, W-sl, di+sf);
	    }

	    g.setColor(Color.red);
	    g.drawString(PTime+"[sec]", 2, fnt);
	}
	public void disp_pickup(Graphics g) {
	    double by = H2/(double)Cby;
	    int l = W/DEVID;
	    if (l < 2)
		l = 2;
	    int r = l/2;

	    disp_cmemori(g);
	    if (K == null)
		return;
	    if (K[0] == null)
		return;

	    int skey = get_key();
	    int pkey = get_pkey(skey);
	    int x0, y0, x1 = 0, y1 = H2;
	    for (int i = 0; i < Loop; i++) {
		if (Pic[i] == null)
		    continue;
		x0 = (int)(dp*i);
		y0 = (int)(H2-(ftoc(pkey, Pic[i].getPeakf())-
			       get_ipcent(skey))*by);
		if (i == 0) {
		    x1 = x0;
		    y1 = y0;
		}
		g.setColor(Psw[i] ? Color.green : Color.gray);
		g.drawLine(x0, y0, x1, y1);
		
		g.setColor(Psw[i] ? Color.white : Color.gray);
		g.fillOval(x0-r, y0-r, l, l);
 		x1 = x0;
		y1 = y0;
	    }

	    disp_keyfc(g, skey, pic_ave());
	}
	public double pic_ave() {
	    double avef = 0;
	    int n = 0;
	    for (int i = 0; i < Loop; i++) {
		if (Pic[i] == null)
		    continue;

		if (Psw[i]) {
		    n++;
		    avef += Pic[i].getPeakf();
		}
	    }

	    return avef/n;
	}
	public void drag_check() {
	    if (Psw == null)
		return;

	    for (int i = 0; i < DEVID; i++) {
		int dx = (int)(dp*i);
		Psw[i] = (dx >= firstX && dx <= lastX) ? true : false;
	    }
	}
    }
    //---------

    public boolean isPeak() {
	for (int i = 0; i < N2; i++) {
	    if (YY[i] > Level)
		return true;
	}

	return false;
    }
    public void reset_peak() {
	K = new Peak[200];
	Peak.resetPeakt();
    }
    public void ifpk() {
	reset_peak();

	int ky, npk = 0;
	double beta, pk, fi;
	for (int i = 2; i < N2-1; i++) {
	    if (YY[i] > Level) {
		if ((Math.abs(YY[i]) >= Math.abs(YY[i-1])) &&
		    (Math.abs(YY[i]) >= Math.abs(YY[i+1]))) {
		    beta = 3.0/(1+YY[i]/YY[i+1])-1;
		    pk = ((Math.PI*(1+beta)*beta*(1-beta))/
			  Math.sin(Math.PI*beta))*YY[i];
		    fi = (i+beta)*DELF;
		    ky = ftok(fi);
		    try {
			K[npk++] = new Peak(ky, fi, pk);
		    } catch (ArrayIndexOutOfBoundsException e) {
			Peak.setPeakt(-1);
			Error = "Peak over 200 error.";
		    }
		    i += 2;
		} 
	    }
	}
    }
    public Peak pick(int num) {
	ifpk();
	//int num = PG.get_select_keynm();

	if (K[num] == null || K[0] == null)
	    return null;

	return K[num];
    }
    public void pickup(int num) {
	if (NOWPK && (Loop < DEVID)) {
	    Pic[Loop] = pick(num);
	    Psw[Loop] = true;

	    if (Loop == DEVID-1 && CUT) //
		System.arraycopy(Pic, 1, Pic, 0, DEVID-1);
	}
    }
    public void reset_pic() {
	Pic = new Peak[DEVID];
	Psw = new boolean[DEVID];
	Loop = 0;
	firstX = 0;
	lastX = PG.W;
    }

    public double ctof(double f, double ce) {
        return f*Math.pow(10, ce*ML);
    }
    public double ftoc(int keyn, double f) {
	return MC*Math.log(f/getFreq(keyn));
    }
    public int ftok(double sf) {
	int key = (int)Math.rint(Oct*log_base(2,sf/getFreq(0)));
        if (key < 0)
            key = 0;
        else if (key >= KC)
            key = KC-1;

        return key;  
    }
    public double log_base(double base, double value) {
        return Math.log(value)/Math.log(base);
    }
    public String dform(double val) {
        return df1.format((float)val);
    }
    public double getFreq(int key) {
        key++;
        return Pitch*Math.pow(2, (key-49)/(double)Oct);
    }
    public void add_cent(double c) {
	for (int i = 0; i < KC; i++)
	    Keyc[i] += c;
    }
    public void initStart() {
	initc(ftoc(49-1, Pitch));
	init_inh(0);
    }
    public void initc(double c) {
	Keyc = new double[KC];

	for (int i = 0; i < KC; i++)
	    Keyc[i] = c;
    }
    public void init_inh(double ih) {
	Keyi = new double[KC];

	for (int i = 0; i < KC; i++)
	    Keyi[i] = ih;
    }
    //---------

    public void initHanning() {
	Han = new double[NN];

	for (int i = 0; i < NN; i++)
	    Han[i] = 0.5-0.5*Math.cos(PI2*i*RN);
    }
    public void hanning(int[] ad) {
	for (int i = 0; i < NN; i++) {
	    Re[i] = ad[i]*Han[i];
	    Im[i] = 0;
	}
    }
    public void fft(int[] n) {
	Re = new double[NN];
	Im = new double[NN];

	hanning(n);
	fftN.fft(Re, Im);
	for (int i = 0; i < N2; i++)
	    YY[i] = Math.sqrt(Re[i]*Re[i]+Im[i]*Im[i]); // Im
    }
}
